import uuid
from fastapi import BackgroundTasks
import requests
from fastapi.encoders import jsonable_encoder
from fastapi.responses import JSONResponse

from colab_easy_ui.Functions import ProgressTaskStatus, ProgressFunctionCallResponse, JsonApiTaskStatus, ProgressTaskProcessStatus
from dataclasses import asdict
import json

import sys
from colab_easy_ui.const import sanitaize_path

sys.path.append("../easy-vc-dev")
from easy_vc_dev.utils.extract_feature import extract_feats as _extract_feats  # noqa


class ExtractFeatsWrapper:
    def __init__(self):
        self.progresses = ProgressTaskStatus(processStatus={})

    def extract_feats_function(self, port: int, uuid_str: str, project_name: str, version: int, device_id: int):
        self.port = port
        self.uuid_str = uuid_str
        try:
            _extract_feats(project_name, version, device_id, self.extract_feats_callback)
        except Exception as e:
            print("extract_feats_function exception:", e)
            open("extract_feats1eeee.txt", "w").write(e)

        self.update_status("DONE")

    def extract_feats_callback(self, data_type: str, n: int, total: int, status: str):
        if data_type not in self.progresses.processStatus:
            self.progresses.processStatus[data_type] = ProgressTaskProcessStatus(
                display_name=data_type,
                n=0,
                total=0,
                status="RUNNING",
                unit="",
            )
        self.progresses.processStatus[data_type].n = n
        self.progresses.processStatus[data_type].total = total
        self.progresses.processStatus[data_type].status = status

        self.update_status("RUNNING")

    def update_status(self, status: JsonApiTaskStatus):
        data = json.dumps(asdict(self.progresses))
        requests.get(f"http://localhost:{self.port}/functions_set_task_status?task_id={self.uuid_str}&status={status}&data={data}")


def extract_feats(port: int, project_name: str, version: int, device_id: int, background_tasks: BackgroundTasks):
    if sanitaize_path(project_name) is False:
        data = ProgressFunctionCallResponse(
            status="NG",
            uuid="",
            message=f"not valid project name {project_name}",
        )
        json_compatible_item_data = jsonable_encoder(data)
        return JSONResponse(content=json_compatible_item_data)

    # UUIDを作成
    uuid_str = str(uuid.uuid4())

    extract_feats_wrapper = ExtractFeatsWrapper()
    background_tasks.add_task(extract_feats_wrapper.extract_feats_function, port, uuid_str, project_name, version, device_id)

    try:
        data = ProgressFunctionCallResponse(
            status="OK",
            uuid=uuid_str,
            message="",
        )
        json_compatible_item_data = jsonable_encoder(data)
        return JSONResponse(content=json_compatible_item_data)
    except Exception as e:
        data = ProgressFunctionCallResponse(
            status="NG",
            uuid="",
            message=str(e),
        )
        json_compatible_item_data = jsonable_encoder(data)
        return JSONResponse(content=json_compatible_item_data)



import zipfile
import os
from typing import Callable
import uuid
from fastapi import BackgroundTasks
import requests
import functools
from fastapi.encoders import jsonable_encoder
from fastapi.responses import JSONResponse

from colab_easy_ui.Functions import ProgressTaskStatus, ProgressFunctionCallResponse, JsonApiTaskStatus, ProgressTaskProcessStatus
from dataclasses import asdict


def exec_unzip(_callback: Callable[[JsonApiTaskStatus, ProgressTaskStatus], None], zip_path: str, extract_to: str):
    progresses = ProgressTaskStatus(processStatus={})

    os.makedirs(extract_to, exist_ok=True)
    progresses.processStatus["unzip"] = ProgressTaskProcessStatus(
        display_name="Unzip",
        n=0,
        total=0,
        status="RUNNING",
        unit="",
    )

    # ZIPファイルを開く
    with zipfile.ZipFile(zip_path, "r") as zip_ref:
        # 解凍するファイルリストを取得
        file_list = zip_ref.namelist()
        # ファイルの総数を取得し、進捗バーの最大値として設定
        total_files = len(file_list)
        # tqdmを使用して進捗表示を行う
        for i, file in enumerate(file_list):
            # ファイルを解凍
            zip_ref.extract(file, extract_to)
            progresses.processStatus["unzip"].n = i
            progresses.processStatus["unzip"].total = total_files
            progresses.processStatus["unzip"].status = "RUNNING"
            _callback("RUNNING", progresses)

        progresses.processStatus["unzip"].status = "DONE"
        _callback("DONE", progresses)


def unzip_callback(status: JsonApiTaskStatus, allStatus: ProgressTaskStatus, port: int, uuid_str: str):
    import json

    data = json.dumps(asdict(allStatus))
    requests.get(f"http://localhost:{port}/functions_set_task_status?task_id={uuid_str}&status={status}&data={data}")


def unzip_function(port: int, uuid_str: str, zip_path: str, extract_to: str):
    unzip_callback_fixed = functools.partial(unzip_callback, port=port, uuid_str=uuid_str)
    exec_unzip(unzip_callback_fixed, zip_path, extract_to)


def unzip(port: int, zip_path: str, extract_to: str, background_tasks: BackgroundTasks):
    # UUIDを作成
    uuid_str = str(uuid.uuid4())

    background_tasks.add_task(unzip_function, port, uuid_str, zip_path, extract_to)

    try:
        data = ProgressFunctionCallResponse(
            status="OK",
            uuid=uuid_str,
            message="",
        )
        json_compatible_item_data = jsonable_encoder(data)
        return JSONResponse(content=json_compatible_item_data)
    except Exception as e:
        data = ProgressFunctionCallResponse(
            status="NG",
            uuid="",
            message=str(e),
        )
        print(data)
        return JSONResponse(content=json_compatible_item_data)



from fastapi.encoders import jsonable_encoder
from fastapi.responses import JSONResponse

from colab_easy_ui.Functions import GetFunctionCallResponse
import os
import json


def get_data_epochs(project_folder: str, project_name: str):
    datafiles = [x for x in os.listdir(os.path.join(project_folder, project_name, "logs")) if os.path.isfile(os.path.join(project_folder, project_name, "logs", x))]
    # epochs = [x.split("-")[2][1:] for x in datafiles if x.split("-")[2].startswith("e")]
    epochs: list[int] = []
    for file in datafiles:
        elem = file.split("-")
        if len(elem) < 3:
            continue
        if not elem[2].startswith("e"):
            continue
        epoch_str = elem[2][1:]
        if not epoch_str.isdecimal():
            continue
        epochs.append(int(epoch_str))

    epochs = sorted(set(epochs))
    return epochs


def get_generator_model_files(project_folder: str, project_name: str):
    datafiles = [x for x in os.listdir(os.path.join(project_folder, project_name, "logs")) if os.path.isfile(os.path.join(project_folder, project_name, "logs", x))]
    generator_model_files = [x for x in datafiles if x.endswith("gen.pt")]
    return generator_model_files


def get_best(project_folder: str, project_name: str):
    best_json = os.path.join(project_folder, project_name, "logs", "best.json")
    if not os.path.isfile(best_json):
        return {}

    best_data = json.loads(open(best_json, "r").read())
    return best_data


def get_server_info_function(
    dataset_folder: str,
    project_folder: str,
):
    if not os.path.isdir(dataset_folder):
        os.makedirs(dataset_folder)
    if not os.path.isdir(project_folder):
        os.makedirs(project_folder)
    datasets = [x for x in os.listdir(dataset_folder) if os.path.isdir(os.path.join(dataset_folder, x))]
    # projects = [x for x in os.listdir(project_folder) if os.path.isdir(os.path.join(project_folder, x))]
    try:
        projects = [
            {
                "name": x,
                "saved_epochs": get_data_epochs(project_folder, x),
                "best": get_best(project_folder, x),
                "generator_model_files": get_generator_model_files(project_folder, x),
            }
            for x in os.listdir(project_folder)
            if os.path.isdir(os.path.join(project_folder, x))
        ]
    except Exception as e:
        print(e)

    try:
        data = GetFunctionCallResponse(
            status="OK",
            message="",
            data={
                "datasets": datasets,
                "projects": projects,
            },
        )
        json_compatible_item_data = jsonable_encoder(data)
        return JSONResponse(content=json_compatible_item_data)
    except Exception as e:
        data = GetFunctionCallResponse(
            status="NG",
            message=str(e),
            data={},
        )
        print(data)
        return JSONResponse(content=json_compatible_item_data)



from fastapi.encoders import jsonable_encoder
from fastapi.responses import JSONResponse
from fastapi import HTTPException
from colab_easy_ui.Functions import ProgressFunctionCallResponse
import os
import sys
from fastapi.responses import FileResponse
from colab_easy_ui.const import is_running_on_colab, sanitaize_path

sys.path.append("../easy-vc-dev")
from easy_vc_dev.onnx.export_onnx import export_onnx as _export_onnx  # noqa


def onnx_download(
    # project_name: str,
    # torch_file: str,
    # ipython=None,
    display=None,
):
    open("onnx_download1.txt", "w").write("Eonnx_download")

    if is_running_on_colab() is True:
        open("onnx_download2.txt", "w").write("ERonnx_download")
        try:
            display.download("/content/sample_data/mnist_train_small.csv")
        except Exception as e:
            open("onnx_download2-2-2.txt", "w").write(str(e))
        open("onnx_download3.txt", "w").write("ERonnx_download")
        # from google.colab import files

        # files.download("/content/sample_data/mnist_train_small.csv")

    return

    # if sanitaize_path(project_name) is False or sanitaize_path(torch_file) is False:
    #     data = ProgressFunctionCallResponse(
    #         status="NG",
    #         uuid="",
    #         message=f"not valid project name {project_name} or torch_file {torch_file}",
    #     )
    #     json_compatible_item_data = jsonable_encoder(data)
    #     return JSONResponse(content=json_compatible_item_data)

    # file_path = os.path.join("trainer", project_name, "logs", os.path.splitext(torch_file)[0] + ".onnx")

    # if not os.path.exists(file_path) or not os.path.isfile(file_path):
    #     raise HTTPException(status_code=404, detail="File not found")

    # file_size = os.path.getsize(file_path)
    # headers = {
    #     "Content-Length": str(file_size),
    # }

    # return FileResponse(path=file_path, filename=os.path.basename(file_path), media_type="application/octet-stream", headers=headers)



from fastapi import BackgroundTasks
from colab_easy_ui.Functions import ProgressFunctionCallResponse, ProgressTaskStatus, JsonApiTaskStatus
from typing import Callable
from dataclasses import asdict
from colab_easy_ui.plugins.download_function.Downloader import DownloadParams, Downloader
import requests
import functools
from fastapi.encoders import jsonable_encoder
from fastapi.responses import JSONResponse
import uuid


def exec_download(callback: Callable[[JsonApiTaskStatus, ProgressTaskStatus], None], downloadParams: list[DownloadParams]):
    downloader = Downloader(callback)
    for p in downloadParams:
        downloader.pushItem(p)
    downloader.download()
    callback("DONE", downloader.progresses)


def download_callback(status: JsonApiTaskStatus, allStatus: ProgressTaskStatus, port: int, uuid_str: str):
    import json

    data = json.dumps(asdict(allStatus))
    requests.get(f"http://localhost:{port}/functions_set_task_status?task_id={uuid_str}&status={status}&data={data}")


def download_function(port: int, uuid_str: str, downloadParams: list[DownloadParams]):
    download_callback_fixed = functools.partial(download_callback, port=port, uuid_str=uuid_str)
    exec_download(download_callback_fixed, downloadParams)


def download(port: int, downloadParams: list[DownloadParams], background_tasks: BackgroundTasks):
    # UUIDを作成
    uuid_str = str(uuid.uuid4())

    background_tasks.add_task(download_function, port, uuid_str, downloadParams)
    # await download_function(background_tasks, port, uuid_str, downloadParams)

    try:
        data = ProgressFunctionCallResponse(
            status="OK",
            uuid=uuid_str,
            message="",
        )
        json_compatible_item_data = jsonable_encoder(data)
        return JSONResponse(content=json_compatible_item_data)
    except Exception as e:
        data = ProgressFunctionCallResponse(
            status="NG",
            uuid="",
            message=str(e),
        )
        print(data)
        return JSONResponse(content=json_compatible_item_data)



import fire
from colab_easy_ui.ColabEasyUI import ColabEasyUI, JsonApiFunc

import functools
from colab_easy_ui.plugins.download_function.Downloader import DownloadParams
from colab_easy_ui.plugins.download_function.download_function import download
from colab_easy_ui.plugins.extract_feats_function import extract_feats
from colab_easy_ui.plugins.generate_filelist_function import generate_filelist
from colab_easy_ui.plugins.get_server_info_function import get_server_info_function
from colab_easy_ui.plugins.onnx_download_function import onnx_download

from colab_easy_ui.plugins.preprocess_function import preprocess
from colab_easy_ui.plugins.train_function import train
from colab_easy_ui.plugins.onnx_export_function import onnx_export

from colab_easy_ui.plugins.unzip_function import unzip
import logging
import os

downloadParams = [
    DownloadParams(
        display_name="whisper_tiny.pt",
        url="https://openaipublic.azureedge.net/main/whisper/models/65147644a518d12f04e32d6f3b26facc3f8dd46e5390956a9424a650c0ce22b9/tiny.pt",
        saveTo="./models/embedder/whisper_tiny.pt",
        hash="65147644a518d12f04e32d6f3b26facc3f8dd46e5390956a9424a650c0ce22b9",
    ),
    DownloadParams(
        display_name="nof0-tsu_e1500.pt",
        url="https://huggingface.co/wok000/vcclient_model/resolve/main/easy-vc/nof0-tsu_e1500.pt",
        saveTo="./models/pretrained/easy-vc/nof0-tsu_e1500.pt",
        hash="ef3a84f26e4fe616bc9adf849b2e86b6117f865e3278473f77feec75550c344b",
    ),
]


logging.getLogger("urllib3").setLevel(logging.WARNING)
logging.getLogger("multipart").setLevel(logging.WARNING)


def run_server(
    ipython=None,
    display=None,
    logfile=None,
):
    c = ColabEasyUI.get_instance(ipython)
    port = c.port
    # ファイルアップローダ
    c.enable_file_uploader("upload", {"abc": "voice.zip"})

    # Tensorboardの登録
    c.enable_colab_internal_fetcher("trainer", ipython, logfile)

    # 機能登録
    c.register_functions(
        [
            # backgroundタスクのパラレル化がむずいので、一つずつ別タスクにする（TOBE IMPROVED）。
            JsonApiFunc("download_whisper", "progress", "whisper", "GET", "/download1", functools.partial(download, port=port, downloadParams=downloadParams[0:1])),
            JsonApiFunc("download_pretrain", "progress", "pretrain", "GET", "/download2", functools.partial(download, port=port, downloadParams=downloadParams[1:2])),
            JsonApiFunc("unzip", "progress", "unzip", "GET", "/unzip", functools.partial(unzip, port=port, zip_path="upload/voice.zip", extract_to="raw_data")),
            JsonApiFunc(
                "preprocess",
                "progress",
                "preprocess",
                "GET",
                "/preprocess",
                functools.partial(
                    preprocess,
                    port=port,
                    # project_name="amitaro2",
                    # wav_folder="raw_data/amitaro2",
                    # sample_rate=16000,
                    # jobs=4,
                    # valid_num=10,
                    # test_folder="raw_data/test",
                ),
            ),
            JsonApiFunc(
                "extract_feats",
                "progress",
                "extract feats",
                "GET",
                "/feats",
                functools.partial(
                    extract_feats,
                    port=port,
                    # project_name="amitaro2",
                    # version=1,
                    # device_id=0,
                ),
            ),
            JsonApiFunc(
                "generate_filelist",
                "progress",
                "generate filelist",
                "GET",
                "/filelist",
                functools.partial(
                    generate_filelist,
                    port=port,
                    # project_name="amitaro2",
                    # version=1,
                    # useF0=False,
                    # sid=0,
                ),
            ),
            JsonApiFunc(
                "start_train",
                "progress",
                "start train",
                "GET",
                "/train",
                functools.partial(
                    train,
                    port=port,
                    # project_name="amitaro2",
                    # config_path="16k_v2.json",
                    # sample_rate=16000,
                    # use_f0=False,
                    # total_epoch=5,
                    # batch_size=10,
                    # device_id=0,
                    # log_step_interval=100,
                    # val_step_interval=100,
                    # test_step_interval=100,
                    # save_model_epoch_interval=1,
                    # # checkpoint_path: str | None = None,
                    # # cache_gpu: bool = False,
                    # # freeze_vocoder: bool = True,
                    # # finetune: bool = False,
                ),
            ),
            JsonApiFunc(
                "onnx_export",
                "progress",
                "onnx export",
                "GET",
                "/onnx_export",
                functools.partial(
                    onnx_export,
                    port=port,
                ),
            ),
            # JsonApiFunc("get_dataset_id", "get", "get_dataset_name", "GET", "/get_dataset", functools.partial(get_dataset_function, dataset_folder="raw_data")),
            JsonApiFunc("get_server_info", "get", "get_server_info", "GET", "/get_server_info", functools.partial(get_server_info_function, dataset_folder="raw_data", project_folder="trainer")),
            JsonApiFunc("onnx_download", "download", "onnx_download", "GET", "/onnx_download", functools.partial(onnx_download, display=display)),
            # JsonApiFunc("onnx_download", "download", "onnx_download", "GET", "/onnx_download", functools.partial(onnx_download, ipython=ipython)),
        ]
    )

    # 静的ファイルのマウント
    current_file_path = os.path.abspath(__file__)
    current_dir = os.path.dirname(current_file_path)
    static_files_path = os.path.join(current_dir, "frontend/dist")
    print(static_files_path)
    c.mount_static_folder("/front2", static_files_path)
    # c.mount_static_folder("/front2", "frontend/dist")

    # サーバー起動
    port = c.start()
    print(port)
    return port


def main():
    fire.Fire(run_server)



from dataclasses import dataclass
from typing import TypeAlias, Literal

from typing import Dict


# from typing import Union, Dict

RESPONSE_STATUS: TypeAlias = Literal[
    "OK",
    "NG",
]


@dataclass
class ColabEasyUIResponse:
    status: RESPONSE_STATUS
    message: str


# @dataclass
# class ColabInternalFetchResponse(Response):
#     url: str
#     data: Dict[str, Union[str, int, float]] | None


@dataclass
class EasyFileUploaderResponse(ColabEasyUIResponse):
    allowed_filenames: Dict[str, str] | None = None



def is_running_on_colab():
    try:
        import google.colab  # NOQA

        return True
    except ImportError:
        return False


def sanitaize_path(path: str) -> bool:
    if path.startswith(".") or path.startswith("/"):
        return False
    return True



import uuid
from fastapi import BackgroundTasks
import requests
from fastapi.encoders import jsonable_encoder
from fastapi.responses import JSONResponse

from colab_easy_ui.Functions import ProgressTaskStatus, ProgressFunctionCallResponse, JsonApiTaskStatus, ProgressTaskProcessStatus
from dataclasses import asdict
import json
import os
import sys
import re
from colab_easy_ui.const import sanitaize_path

sys.path.append("../easy-vc-dev")
from easy_vc_dev.train.train import train as _train  # noqa


class TrainWrapper:
    def __init__(self):
        self.progresses = ProgressTaskStatus(processStatus={})

    def train_function(
        self,
        port: int,
        uuid_str: str,
        project_name: str,
        # config_path: str,
        sample_rate: int,
        use_f0: bool,
        total_epoch: int,
        batch_size: int,
        device_id: int,
        log_step_interval: int,
        val_step_interval: int,
        test_step_interval: int,
        save_model_epoch_interval: int,
        checkpoint_path: str | None = None,
        cache_gpu: bool = False,
        freeze_vocoder: bool = False,
        finetune: bool = True,
        vocoder_ckpt: str | None = None,
    ):
        self.port = port
        self.uuid_str = uuid_str
        try:
            _train(
                project_name,
                # config_path,
                sample_rate,
                use_f0,
                total_epoch,
                batch_size,
                device_id,
                log_step_interval,
                val_step_interval,
                test_step_interval,
                save_model_epoch_interval,
                checkpoint_path,
                cache_gpu,
                freeze_vocoder,
                finetune,
                vocoder_ckpt,
                self.train_callback,
            )
        except Exception as e:
            print(e)
            open("train_exception.txt", "w").write(str(e))

        self.update_status("DONE")

    def train_callback(self, data_type: str, n: int, total: int, status: str):
        if data_type not in self.progresses.processStatus:
            self.progresses.processStatus[data_type] = ProgressTaskProcessStatus(
                display_name=data_type,
                n=0,
                total=0,
                status="RUNNING",
                unit="",
            )
        self.progresses.processStatus[data_type].n = n
        self.progresses.processStatus[data_type].total = total
        self.progresses.processStatus[data_type].status = status

        latest_keys = sorted(self.progresses.processStatus.keys())[-3:]
        new_array = {}
        for key in latest_keys:
            new_array[key] = self.progresses.processStatus[key]

        self.progresses.processStatus = new_array

        self.update_status("RUNNING")

    def update_status(self, status: JsonApiTaskStatus):
        data = json.dumps(asdict(self.progresses))
        requests.get(f"http://localhost:{self.port}/functions_set_task_status?task_id={self.uuid_str}&status={status}&data={data}")


def train(
    port: int,
    project_name: str,
    # config_path: str,
    sample_rate: int,
    use_f0: bool,
    total_epoch: int,
    batch_size: int,
    device_id: int,
    log_step_interval: int,
    val_step_interval: int,
    test_step_interval: int,
    save_model_epoch_interval: int,
    background_tasks: BackgroundTasks,
    # checkpoint_path: str | None = None,
    resume_from_epoch: int = 0,
    cache_gpu: bool = False,
    freeze_vocoder: bool = False,
    # finetune: bool = False,
):
    if sanitaize_path(project_name) == False:
        data = ProgressFunctionCallResponse(
            status="NG",
            uuid="",
            message=f"not valid project name {project_name}",
        )
        json_compatible_item_data = jsonable_encoder(data)
        return JSONResponse(content=json_compatible_item_data)

    # UUIDを作成
    uuid_str = str(uuid.uuid4())

    if resume_from_epoch == 0:
        print("train from pretrained.")
        checkpoint_path = "models/pretrained/easy-vc/nof0-tsu_e1500.pt"
        finetune = True
    else:
        print(f"resume training from epoch {resume_from_epoch}.")
        data_dir = os.path.join("trainer", project_name, "logs")
        datafiles = [x for x in os.listdir(data_dir) if os.path.isfile(os.path.join(data_dir, x))]

        epoch_pattern = rf"^.+-e{resume_from_epoch:04d}-.+\d+\.pt$"
        matched_files = [f for f in datafiles if re.match(epoch_pattern, f)]
        assert len(matched_files) == 1
        checkpoint_path = os.path.join(data_dir, matched_files[0])
        finetune = False

    train_wrapper = TrainWrapper()
    background_tasks.add_task(
        train_wrapper.train_function,
        port,
        uuid_str,
        project_name,
        # config_path,
        sample_rate,
        use_f0,
        total_epoch,
        batch_size,
        device_id,
        log_step_interval,
        val_step_interval,
        test_step_interval,
        save_model_epoch_interval,
        checkpoint_path,
        cache_gpu,
        freeze_vocoder,
        finetune,
    )

    try:
        data = ProgressFunctionCallResponse(
            status="OK",
            uuid=uuid_str,
            message="",
        )
        json_compatible_item_data = jsonable_encoder(data)
        return JSONResponse(content=json_compatible_item_data)
    except Exception as e:
        data = ProgressFunctionCallResponse(
            status="NG",
            uuid="",
            message=str(e),
        )
        print(data)
        return JSONResponse(content=json_compatible_item_data)



import uuid
from fastapi import BackgroundTasks
import requests
from fastapi.encoders import jsonable_encoder
from fastapi.responses import JSONResponse

from colab_easy_ui.Functions import ProgressTaskStatus, ProgressFunctionCallResponse, JsonApiTaskStatus, ProgressTaskProcessStatus
from dataclasses import asdict
import json
import os
import sys

sys.path.append("../easy-vc-dev")
from easy_vc_dev.onnx.export_onnx import export_onnx as _export_onnx  # noqa


class OnnxExportWrapper:
    def __init__(self):
        self.progresses = ProgressTaskStatus(processStatus={})

    def onnx_export_function(
        self,
        port: int,
        uuid_str: str,
        torch_path: str,
        output_path: str,
        static_size: int = 0,
    ):
        self.port = port
        self.uuid_str = uuid_str

        try:
            _export_onnx(
                torch_path,
                output_path,
                static_size,
                self.export_callback,
            )
        except Exception as e:
            print(e)

        self.update_status("DONE")

    def export_callback(self, data_type: str, n: int, total: int, status: str):
        if data_type not in self.progresses.processStatus:
            self.progresses.processStatus[data_type] = ProgressTaskProcessStatus(
                display_name=data_type,
                n=0,
                total=0,
                status="RUNNING",
                unit="",
            )
        self.progresses.processStatus[data_type].n = n
        self.progresses.processStatus[data_type].total = total
        self.progresses.processStatus[data_type].status = status
        self.update_status("RUNNING")

    def update_status(self, status: JsonApiTaskStatus):
        data = json.dumps(asdict(self.progresses))
        requests.get(f"http://localhost:{self.port}/functions_set_task_status?task_id={self.uuid_str}&status={status}&data={data}")


def onnx_export(
    port: int,
    project_name: str,
    torch_file: str,
    background_tasks: BackgroundTasks,
    static_size: int = 0,
):
    # UUIDを作成
    uuid_str = str(uuid.uuid4())
    torch_path = os.path.join("trainer", project_name, "logs", torch_file)
    output_path = os.path.join("trainer", project_name, "logs", os.path.splitext(torch_file)[0] + ".onnx")

    onnx_export_wrapper = OnnxExportWrapper()

    background_tasks.add_task(
        onnx_export_wrapper.onnx_export_function,
        port,
        uuid_str,
        torch_path,
        output_path,
        static_size,
    )

    try:
        data = ProgressFunctionCallResponse(
            status="OK",
            uuid=uuid_str,
            message="",
        )
        json_compatible_item_data = jsonable_encoder(data)
        return JSONResponse(content=json_compatible_item_data)
    except Exception as e:
        data = ProgressFunctionCallResponse(
            status="NG",
            uuid="",
            message=str(e),
        )
        print(data)
        return JSONResponse(content=json_compatible_item_data)



from dataclasses import dataclass
import os
import requests
import hashlib
from typing import Callable


from colab_easy_ui.Functions import JsonApiTaskProcessStatus, JsonApiTaskStatus, ProgressTaskProcessStatus, ProgressTaskStatus
from colab_easy_ui.const import is_running_on_colab


if is_running_on_colab():
    print("load tqdm for notebook")
    from tqdm.notebook import tqdm
else:
    print("load normal notebook")
    from tqdm import tqdm


@dataclass
class DownloadParams:
    display_name: str
    url: str
    saveTo: str
    hash: str | None
    position: int | None = None


@dataclass
class HashCheckResult:
    display_name: str
    url: str
    status: str
    message: str


class Downloader:
    def __init__(self, callback: (Callable[[JsonApiTaskStatus, ProgressTaskStatus], None]) | None = None):
        self.callback = callback
        self.params: list[DownloadParams] = []
        self.progresses = ProgressTaskStatus(processStatus={})

    # def pushItem(self, display_name: str, url: str, saveTo: str, hash: str | None):
    #     self.params.append(DownloadParams(display_name, url, saveTo, hash, len(self.params)))

    def pushItem(self, param: DownloadParams):
        position = len(self.params)
        param.position = position
        self.params.append(param)

    def updateProgress(self, display_name: str, url: str, n: int, total: int, status: JsonApiTaskProcessStatus):
        self.progresses.processStatus[display_name] = ProgressTaskProcessStatus(
            display_name=display_name,
            n=n if n >= 0 else self.progresses.processStatus[display_name].n,
            total=total if total >= 0 else self.progresses.processStatus[display_name].total,
            status=status,
            unit="byte",
        )
        if self.callback is not None:
            self.callback("RUNNING", self.progresses)

    def download(self):
        for p in self.params:
            self._download_and_check(p)
        # 以下、並列化を試行錯誤したが、うまくいかなかった残骸。 -> Threadingではなくてmultiprocessing.Poolでならいける説。要検証
        # import asyncio

        # tasks = [self._download_and_check(p) for p in self.params]
        # await asyncio.gather(*tasks)

        # with ThreadPoolExecutor() as pool:
        #     pool.map(self._downloadItem, self.params)
        # sleep(1)

        # for p in self.params:
        #     self.background_tasks.add_task(self._download_and_check, p)
        #     print(f"add task{p.saveTo}")
        #     # self._downloadItem(p)

        # import asyncio

        # # loop = asyncio.get_event_loop()
        # loop = asyncio.get_running_loop()
        # futures = [
        #     loop.run_in_executor(None, self._downloadItem, param)  # Noneはデフォルトのexecutorを使うことを意味します。
        #     # カスタムのexecutorを使う場合は、Noneの代わりにそれを指定します。
        #     for param in self.params
        # ]

        # # 全てのfutureが完了するのを待ちます
        # for future in asyncio.as_completed(futures):
        #     await future  # イテレートすることで、各ダウンロードの完了を待ちます。

    def _download_and_check(self, params: DownloadParams):
        self._downloadItem(params)
        self.check_hash(params)

    def _downloadItem(self, params: DownloadParams):
        display_name = params.display_name
        url = params.url
        saveTo = params.saveTo
        position = params.position
        dirname = os.path.dirname(saveTo)
        if dirname != "":
            os.makedirs(dirname, exist_ok=True)

        try:
            req = requests.get(url, stream=True, allow_redirects=True)

            content_length_header = req.headers.get("content-length")
            content_length = int(content_length_header) if content_length_header is not None else 1024 * 1024 * 1024

            progress_bar = tqdm(
                total=content_length if content_length > 0 else None,
                leave=False,
                unit="B",
                unit_scale=True,
                unit_divisor=1024,
                position=position,
            )

            # with tqdm
            chunk_num = content_length // 1024
            callbacl_interval = chunk_num // 100
            with open(saveTo, "wb") as f:
                for i, chunk in enumerate(req.iter_content(chunk_size=1024)):
                    if chunk:
                        progress_bar.update(len(chunk))

                        if i % callbacl_interval == 0:
                            self.updateProgress(display_name, url, progress_bar.n, progress_bar.total, "RUNNING")
                            # print(f"{saveTo}, {i}/{chunk_num}")
                        f.write(chunk)

            self.updateProgress(display_name, url, progress_bar.n, progress_bar.total, "VALIDATING")

        except Exception as e:
            print(e)

    def check_hash(self, params: DownloadParams):
        if not os.path.exists(params.saveTo):
            self.updateProgress(params.display_name, params.url, -1, -1, "ERROR_DOWNLOAD_NOT_FOUND")

        with open(params.saveTo, "rb") as f:
            data = f.read()
            hash = hashlib.sha256(data).hexdigest()
            if hash != params.hash:
                self.updateProgress(params.display_name, params.url, -1, -1, "ERROR_INVALID_CHECKSUM")
                print(f"hash check failed: hash:{hash}, tested:{params.hash}")
            else:
                self.updateProgress(params.display_name, params.url, -1, -1, "DONE")

    # def check_hash(self):
    #     for target in self.params:
    #         if not os.path.exists(target.saveTo):
    #             self.updateProgress(target.display_name, target.url, -1, -1, "ERROR_DOWNLOAD_NOT_FOUND")
    #             continue

    #         with open(target.saveTo, "rb") as f:
    #             data = f.read()
    #             hash = hashlib.sha256(data).hexdigest()
    #             if hash != target.hash:
    #                 self.updateProgress(target.display_name, target.url, -1, -1, "ERROR_INVALID_CHECKSUM")
    #             else:
    #                 self.updateProgress(target.display_name, target.url, -1, -1, "DONE")
